/*
 * gexiv2-log-private.h
 *
 * Author(s)
 * Jim Nelson <jim@yorba.org>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef GEXIV2_LOG_PRIVATE_H
#define GEXIV2_LOG_PRIVATE_H

#include <gexiv2/gexiv2-log.h>
#include <exiv2/error.hpp>

G_BEGIN_DECLS

G_GNUC_INTERNAL gboolean        gexiv2_log_is_handler_installed(void);

G_END_DECLS

#endif /* GEXIV2_LOG_PRIVATE_H */
